## tpad version 1.3
# WinXP (TM) enhanced Notepad clone written in Tcl/Tk.
# Copyright (C) 2004  Antonio Bonifati
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details:
#	<http://www.fsf.org/copyleft/gpl.html>
#
# Requires at least Tcl/Tk 8.4.
# Please report any bug found to the author:
#
#	Antonio Bonifati <ant@venus.deis.unical.it>
#	http://monitor.deis.unical.it/ant
#	Snail mail:
#		Via Ernesto Moneta, 11
#		87012 Castrovillari (CS)
#		Italy
#
## last revised on: $Date: 2004/08/31 $

package provide tpad 1.3

package require Tcl 8.4
package require Tk 8.4      ;# uses text(n) undo mechanism
package require msgcat 1.3  ;# uses message catalog

namespace eval ::tpad {
  # init commands: note that the tpad command must be called
  # *before* any other exported command
  namespace export tpad display

  # public editor menu commands
  namespace export new next tpadOpen save saveAs printSetup print quit \
    undoRedo undo redo cut copy paste tpadDelete findReplace findNext \
    goTo selectAll insertDate insertFile toggleWordWrap ioEncodings \
    toggleStatusBar toggleAsciiTable toggleHtmlTidy toggleTextFormatter \
    toggleXmlwf help tip about

  # public useful and safe commands (not menu commands)
  namespace export lineNo lines goToLine updateStatusBar whatIs cmdHelp \
    emptyOpenRecent openRecentLength getInputEnc getOutputEnc
}

## private commands
proc ::tpad::LoadConf {} {
  set global /usr/local/etc/tpad.conf

  set peruser ~/.tpadrc

  if { [file exists $peruser] } {
    set configFile [Source $peruser]
  } elseif { [file exists $global] } {
    set configFile [Source $global]
  } else {
    puts [format [mc "Cannot find a global or per-user config file.\nPlease put either %1\$s\nor %2\$s in place."] $global $peruser]
    exit 1
  }

  # check the configuration file version
  if {![info exists ::tpad::version] || $::tpad::version!={1.3}} {
    puts [format [mc "The configuration file %s\ndoesn't seem to belong to this version of tpad.\nTo avoid any problems, please make sure to use a\nconfig file that exactly matches this version (1.3).\n\nPlease refer to \"Configuration / Upgrading\" in tpad documentation."] $configFile $::tpad::version]
    exit 1
  }
}

proc ::tpad::Source {fileName} {
  if { [catch {source $fileName}] } {
    puts [format [mc "Error in config file:\n\n%s" $::errorInfo]]
    exit 1
  }
  return $fileName
}

# (Slightly modified version of ::tk::FindAltKeyTarget)
# search recursively through the hierarchy of visible widgets
# to find button or label which has $char as underlined character
#
proc ::tpad::FindAltKeyTarget {path char} {
  switch -glob [winfo class $path] {
    *[bB]utton -
    Label {
      if {[string equal -nocase $char [string index [$path cget -text] \
            [$path cget -underline]]]} {return $path} else {return {}}
    }

    default {
      if [catch {$path cget -labelwidget} lw] {set lw ""}

      foreach child [concat $lw [grid slaves $path] \
          [pack slaves $path] [place slaves $path] ] {
        if {""!=[set target [::tpad::FindAltKeyTarget $child $char]]} {
          return $target
        }
      }
    }
  }
  return {}
}

# (Slightly modified version of ::tk::AltKeyInDialog)
# <Alt-Key> event handler for standard dialogs. Sends <<AltUnderlined>>
# to button or label which has appropriate underlined character
#
proc ::tpad::AltKeyInDialog {path key} {
  set target [FindAltKeyTarget $path $key]
  if { $target==""} return
  event generate $target <<AltUnderlined>>
}

# (Slightly modified version of ::tk::AmpWidget)
# Creates new widget, turning -text option into -text and
# -underline options, returned by ::tk::UnderlineAmpersand.
#
proc ::tpad::AmpWidget {class path args} {
  set wcmd [list $class $path]
  foreach {opt val} $args {
    if {[string equal $opt {-text}]} {
      foreach {newtext under} [::tk::UnderlineAmpersand $val] {
        lappend wcmd -text $newtext -underline $under
      }
    } else {
      lappend wcmd $opt $val
    }
  }
  eval $wcmd
  if {[string match *button $class]} {
    bind $path <<AltUnderlined>> [list $path invoke]
  }
  return $path
}

# similar to AmpWidget, but specific for menu entries
proc ::tpad::AmpMenu {path cmd subcmd args} {
   set wcmd [list $path $cmd $subcmd]
   foreach {opt val} $args {
     if {[string equal $opt {-lab}]} {
       foreach {newtext under} [::tk::UnderlineAmpersand $val] {
         lappend wcmd -lab $newtext -underline $under
       }
     } else {
       lappend wcmd $opt $val
     }
   }
   eval $wcmd 
}

proc ::tpad::TextWOampersand {text} {
  return [lindex [::tk::UnderlineAmpersand [mc $text]] 0]
}

# enable/disable edit menu commands based on selection
proc ::tpad::SetStateFromSelection {pn menu} {
  variable $pn\::p ;# also ${pn}::p is ok
  variable $pn\::edit

  set copy [TextWOampersand C&opy]
  set cut [TextWOampersand &Cut]
  set delete [TextWOampersand &Delete]
  if { [catch {$p.text index sel.first}] } {
    $menu entryconf $copy -state disabled
    if {$edit} {
      $menu entryconf $cut -state disabled
      $menu entryconf $delete -state disabled
    }
  } else {
    $menu entryconf $copy -state normal
    if {$edit} {
      $menu entryconf $cut -state normal
      $menu entryconf $delete -state normal
    }
  }

  if {$edit && $::tpad::fmtkind!={none}} {
    variable $pn\::format

    if {$format} {${::tpad::fmtkind}fmt::AutosetRange $pn}
  }
}

proc ::tpad::SetMenuOptsState {pn} {
  variable $pn\::p

  SetStateFromSelection $pn $p.menubar.edit
  SetStateFromSelection $pn $p.rightmenu
}

# enable/disable edit menu commands based on clipboard state
proc ::tpad::SetStateFromClipboard {pn menu} {
  variable $pn\::edit

  set paste [TextWOampersand &Paste]
  if {$edit && !$::tpad::tearoff} {
    if { [catch {selection get -selection CLIPBOARD}] } {
      $menu entryconf $paste -state disabled
    } else {
      $menu entryconf $paste -state normal
    }
  }
}

# called every time the modified flag changes state
# sets the title bar and
# enables the undo/redo command if modified becomes true
proc ::tpad::Modified {pn} {
  foreach v {p filename edit argv0 writable} {
    variable $pn\::$v
  }

  set displayfn [expr {$filename=={} ? [mc Untitled] : $filename}]
  set star {}
  if {$edit} {
    if { [$p.text edit modified] } {
      set star *
      $p.statusBar.modified configure -text [mc Modified]
      set undo [TextWOampersand &Undo]
      $p.menubar.edit entryconf $undo -state normal
      $p.rightmenu entryconf $undo -state normal
      if {$::tpad::maxundo!=1} {
        set redo [TextWOampersand &Redo]
        $p.menubar.edit entryconf $redo -state normal
        $p.rightmenu entryconf $redo -state normal
      }
    } elseif {$writable} {
      $p.statusBar.modified configure -text {}
    } else {
      $p.statusBar.modified configure -text [mc Readonly]
    }
  } else {
    $p.statusBar.modified configure -text [mc Readonly]
  }
  wm title $pn [format $::tpad::titlefmt [file tail $argv0] $displayfn$star]
  $p.statusBar.filename configure -text $displayfn
}

# clears the undo/redo stacks
# sets text modified state as modified (default no)
# and disables the undo/redo command
# sets the insertion cursor at the beginning
# or the end of the text (defaults to the beginning)
proc ::tpad::Init {pn {modified {no}} {end {no}}} {
  variable $pn\::p
  variable $pn\::edit

  $p.text edit reset
  if {$edit} {
    $p.text edit modified $modified
    set undo [TextWOampersand &Undo]
    $p.menubar.edit entryconf $undo -state disabled
    $p.rightmenu entryconf $undo -state disabled
    if {$::tpad::maxundo!=1} {
      set redo [TextWOampersand &Redo]
      $p.menubar.edit entryconf $redo -state disabled
      $p.rightmenu entryconf $redo -state disabled
    }
  } else {
    $p.text edit modified no
  }

  if {$end} {
    $p.text see end
  } else {
    $p.text mark set insert 1.0
  }

  updateStatusBar $pn
}

# arranges for a new text to be edited
proc ::tpad::NewFile {pn {newFilename {}}} {
  foreach v {p edit filename var writable} {
    variable $pn\::$v
  }

  # sets the current name of the file to edit
  # (use an empty string if not defined yet)
  set filename $newFilename
  set writable yes
  $p.text configure -state normal ;# added to support -R option
  $p.text delete 1.0 end ;# clears the text
  if {$var!={}} {
    upvar #0 $var str
    $p.text insert 1.0 $str
  }
  Init $pn
  if {!$edit} {
    $p.text configure -state disabled
  }
}

# opens the file if it can and arranges for it to be
# edited for the first time otherwise acts as NewFile
# also implements a log file facility Notepad has
# return 1 if the file was opened or created ad new
# 0 if the user cancelled the operation
proc ::tpad::OpenFile {pn newFilename} {
  foreach v {p edit filename var writable} {
    variable $pn\::$v
  }

  if { [catch {open $newFilename} fileid] } {
    if {!$edit} {
      tk_messageBox -title tpad -icon error -message $fileid -parent $pn
      return 0
    } elseif { [tk_messageBox -title tpad -type yesno -icon error -parent $pn \
        -message [format [mc "%s\n\nCreate as new?"] $fileid]]=={yes} } {
      NewFile $pn $newFilename
    } else {
      return 0
    }
  } else {
    set writable [file writable $newFilename]
    fconfigure $fileid -encoding [getInputEnc $pn]

    $p.text configure -state normal ;# added to support -R option

    set filename $newFilename
    $p.text delete 1.0 end

    gets $fileid firstline
    $p.text insert 1.0 "$firstline\n"
    $p.text insert insert [read $fileid]
    if {[$p.text get {insert - 1 char}]=="\n"} {
      $p.text delete {insert - 1 char}
    }
    close $fileid

    if {$::tpad::logfac && $firstline=={.LOG}} {
      if {$edit} {
        $p.text insert insert "\n\n"
        insertDate $pn
        $p.text insert insert "\n"
      }

      Init $pn yes yes
    } else {
      Init $pn [expr {$var!={}}]
    }

    if {!$edit} {
      $p.text configure -state disabled
    }

    AddOpenRecent $pn $newFilename
  }
  return 1
}

# save the editing so far done
# without clearing the undo/redo stacks
# also don't include the dummy last line of the text
proc ::tpad::SaveFile {pn} {
  foreach v {p filename var writable} {
    variable $pn\::$v 
  }

  set fileid [open $filename w]
  fconfigure $fileid -encoding [getOutputEnc $pn]
  puts $fileid [$p.text get 1.0 {end - 1 char}]
  close $fileid
  set writable yes
  if {$var=={}} {
    $p.text edit modified no
  }
}

proc ::tpad::SaveVar {pn} {
  variable $pn\::p
  variable $pn\::var

  upvar #0 $var str
  set str [$p.text get 1.0 {end - 1 char}]
  $p.text edit modified no
}

# provides the user with a last change to save his changes
proc ::tpad::OkToContinue {pn} {
  variable $pn\::p

  if { [$p.text edit modified] } {
    switch [tk_messageBox -type yesnocancel \
        -icon warning -title tpad -parent $pn \
        -message [mc "Text modified.\n\nDo you wish to save your changes?"]] {
      yes {return [save $pn]}

      no {return yes}

      cancel {return no}
    }
  }
  return yes
}

# find the next occurrence of the pattern in the text
# starting from the position of the insertion cursor
# returns 1 if the pattern was found, 0 otherwise
proc ::tpad::FindPattern {pn {noMatchMsg yes}} {
  foreach v {p dir case regexp pattern} {
    variable $pn\::$v
  }

  # convert option names and compute search starting and ending point
  switch $dir {
    up {
      set searchdir -backwards
      if { [catch {$p.text index {sel.first - 1 char}} start] } {
        set start insert
      }
      set stop 1.0
    }

    down {
      set searchdir -forwards
      set start insert
      set stop end
    }
  }
  switch $case no {set searchcase -nocase} yes {set searchcase -exact}
  switch $regexp no {set searchtype -exact} yes {set searchtype -regexp}

  if { [catch {$p.text search $searchdir $searchcase $searchtype \
                 -count len -- $pattern $start $stop} matchIdx] } {
    tk_messageBox -icon error -title tpad -parent $pn -message $matchIdx

    return 0
  } elseif {$matchIdx!={}} {
    set last [$p.text index "$matchIdx + $len chars"]
    $p.text tag remove sel 1.0 end
    $p.text tag add sel $matchIdx $last
    $p.text mark set insert $last
    $p.text see $last
    $p.text see $matchIdx
    updateStatusBar $pn

    return 1
  } else {
    if {$noMatchMsg} {
      tk_messageBox -icon info -title tpad -parent $pn \
        -message [format [mc "Cannot find \"%s\""] $pattern]
    }

    return 0
  }
}

# if there is an occurrence of the pattern and is selected,
# replace it with the substitution string, otherwise do nothing;
# then find the next occurrence (if there's one)
proc ::tpad::ReplacePattern {pn {noMatchMsg yes}} {
  foreach v {p pattern substr case regexp} {
    variable $pn\::$v
  }

  if { ![catch {$p.text index sel.first}] } {
    set first [$p.text index sel.first]
    set last [$p.text index sel.last]
    set sel [$p.text get $first $last]

    switch $case no {set searchcase -nocase} yes {set searchcase {}}
    if {$regexp} {
      set equal [eval [list regexp $searchcase $pattern $sel]]
    } else {
      set equal [eval [list string equal $searchcase $pattern $sel]]
    }

    if {$equal} {
      $p.text configure -autosep no
      $p.text edit separator

      $p.text delete $first $last
      $p.text insert insert $substr
      set len [string length $substr]
      set last [$p.text index "$first + $len chars"]
      $p.text tag add sel $first $last

      $p.text edit separator
      $p.text configure -autosep yes
    }
  }

  return [FindPattern $pn $noMatchMsg]
}

proc ::tpad::ReplaceAll {pn} {
  set ret [ReplacePattern $pn]
  while { $ret } {
    update idletasks ;# animated!
    set ret [ReplacePattern $pn no]
  }
}

proc ::tpad::DuplicateSelection {pn args} {
  variable $pn\::p

  $p.text tag remove dupsel 1.0 end

  if { ![catch {$p.text index sel.first}] } {
    eval $p.text tag add dupsel [$p.text tag ranges sel]
  }
}

proc ::tpad::UpdateOffset {pn first last} {
  variable $pn\::p

  $p.vscroll set $first $last
  $p.statusBar.fraction configure \
    -text [expr round($first*100)]-[expr round($last*100)]%
}

proc ::tpad::Quit {pn} {
  SaveOpenRecent $pn
  destroy $pn
  namespace delete $pn
}

proc ::tpad::LoadOpenRecent {pn} {
  variable $pn\::p

  if {$::tpad::maxopenrec!=0 && $::tpad::openrecfile!={}} {
    set fileid [open $::tpad::openrecfile {RDONLY CREAT}]
    gets $fileid orlist
    close $fileid

    foreach filename $orlist {
      $p.menubar.file.recent add comm -lab $filename \
        -comm [namespace code [list tpadOpen $pn $filename]]
    }

    return [llength $orlist]
  }
  return 0
}

proc ::tpad::LastOpenRecent {pn} {
  variable $pn\::p

  return [expr [$p.menubar.file.recent index last] - 2]
}

proc ::tpad::SaveOpenRecent {pn} {
  variable $pn\::p

  if {$::tpad::maxopenrec!=0 && $::tpad::openrecfile!={}} {
    set last [LastOpenRecent $pn]

    set orlist {}
    for {set i [expr {$::tpad::tearoff ? 1 : 0}]} {$i <= $last} {incr i} {
      set filename [$p.menubar.file.recent entrycget $i -lab]
      lappend orlist $filename
    }

    set fileid [open $::tpad::openrecfile w]
    puts $fileid $orlist
    close $fileid
  }
}

# add newFilename to the list of the most recently opened files
# if it is already there, move it to the top
# keep no more than maxopenrec into the list
proc ::tpad::AddOpenRecent {pn newFilename} {
  variable $pn\::p

  if {$::tpad::maxopenrec==0} return

  set last [LastOpenRecent $pn]
  for {set i [expr {$::tpad::tearoff ? 1 : 0}]} {$i <= $last} {incr i} {
    set label [$p.menubar.file.recent entrycget $i -lab]

    if {$label==$newFilename} {
      $p.menubar.file.recent delete $i
      break
    }
  }

  $p.menubar.file.recent insert 0 comm -lab $newFilename \
    -comm [namespace code "tpadOpen $pn $newFilename"]
  $p.menubar.file.recent entryconf end -state normal

  set len [openRecentLength $pn]
  if {$len > $::tpad::maxopenrec} {
    $p.menubar.file.recent delete [LastOpenRecent $pn]
  }
}

proc ::tpad::PlaceWindow {pn win} {
  switch $::tpad::placing {
    {} {::tk::PlaceWindow $win {}}
    pointer {::tk::PlaceWindow $win pointer center}
    widget {::tk::PlaceWindow $win widget $pn}
  }
}
## end of private commands

## public useful and safe commands (not menu commands)
proc ::tpad::getInputEnc {pn} {
  variable $pn\::inputEnc

  return [expr {$inputEnc=={system} ? [encoding system] : $inputEnc}]
}

proc ::tpad::getOutputEnc {pn} {
  variable $pn\::outputEnc

  return [expr {$outputEnc=={system} ? [encoding system] : $outputEnc}]
}

# returns the line number corresponding to index
proc ::tpad::lineNo {pn index} {
  variable $pn\::p

  return [lindex [split [$p.text index $index] .] 0]
}

# returns the total number of lines in the text
proc ::tpad::lines {pn} {
  return [lineNo $pn {end - 1 char}]
}

# if num is a valid line number in the text
# sets the insertion cursor to the specified
# row:col position (col defaults to 0)
# and adjusts the view
proc ::tpad::goToLine {pn row {col 0}} {
  variable $pn\::p

  set lines [lines $pn]
  set row [string trim $row]

  if {[string is integer -strict $row] && $row>=1 && $row<=$lines} {
    $p.text mark set insert $row.$col
    $p.text see $row.$col
    updateStatusBar $pn
    return yes
  } else {
    tk_messageBox -title [mc "Goto line - tpad"] -icon error -parent $pn \
      -message [format [mc \
      "Illegal line number `%1\$s'\nMust be from 1 to %2\$d"] $row $lines]
    return no
  }
}

proc ::tpad::updateStatusBar {pn} {
  variable $pn\::p

  set coords [split [$p.text index insert] .]
  set row [lindex $coords 0]
  set col [expr [lindex $coords 1] + 1]

  $p.statusBar.coords configure -text $row,$col
  $p.statusBar.lines configure -text [lines $pn]
}

proc ::tpad::whatIs {} {
  return [format [mc "tpad Version %s\nWinXP enhanced\nNotepad clone"] \
    $::tpad::version]
}

proc ::tpad::cmdHelp {} {
  puts [format [mc "\n%1\$s\n\nusage: %2\$s"] [whatIs] \
    {tpad [-Rh] [+#] [file ...]}]
}
## end of public useful commands (not menu commands)

## public editor menu commands
proc ::tpad::new {pn} {
   variable $pn\::var

   if {$var!={} || [OkToContinue $pn]} {
     NewFile $pn
   }
}

# arranges for editing the next file in sequence if there's one
# otherwise simply disables the Next option
# returns 1 if a next file in sequence is actually loaded or created as new
proc ::tpad::next {pn} {
  foreach v {p next argc argv var} {
    variable $pn\::$v
  }

  set nextEntry [TextWOampersand Ne&xt]

  if {$next<$argc} {
    if {$var!={} || [OkToContinue $pn]} {
      set file [file normalize [lindex $argv $next]]
      set ret [OpenFile $pn $file]
      if {[incr next]==$argc} {
        # disable the Next option in advance
        $p.menubar.file entryconf $nextEntry -state disabled
      }

      return $ret
    }
  } else {
    $p.menubar.file entryconf $nextEntry -state disabled
  }

  return 0
}

proc ::tpad::tpadOpen {pn {newFilename {}}} {
  variable $pn\::var

  if {$var!={} || [OkToContinue $pn]} {
    if {$newFilename=={}} {
      set newFilename [tk_getOpenFile -title [mc "Open - tpad"] -parent $pn \
        -filetypes $::tpad::types]
    }

    if {$newFilename!={}} {
      OpenFile $pn $newFilename
    }
  }
}

# if a filename has been defined acts as SaveFile and always returns yes
# if no filename has been defined acts as saveAs
# this proc no more does nothing if the text hasn't been modified
# from last saved version, because one may want to save the text
# to convert it to another encoding, without modifying it
# and also because these optimizations are automatic in modern OS's
proc ::tpad::save {pn} {
  foreach v {p filename var} {
    variable $pn\::$v
  }

  if {$var!={}} {
    if { [$p.text edit modified] } {
      SaveVar $pn
    }
    return yes
  } elseif {$filename!={}} {
    SaveFile $pn
    return yes
  } else {
    return [saveAs $pn]
  }
}

# save the text on a file with a new name
# returns yes if the user has chosen to save
# no if the user has cancelled the operation
proc ::tpad::saveAs {pn} {
  variable $pn\::filename

  set newFilename [tk_getSaveFile -title [mc "Save as - tpad"] -parent $pn \
    -initialfile [mc Untitled] -filetypes $::tpad::types]

  if {$newFilename!={}} {
    set filename $newFilename
    SaveFile $pn
    return yes
  }
  return no
}

proc ::tpad::ioEncodings {pn} {
  variable $pn\::p
  variable $pn\::dummy

  set sysEnc [encoding system]

  set m $p.enc
  toplevel $m
  wm group $m $pn
  wm title $m [mc "I/O Encodings - tpad"]
  wm resizable $m no no
  wm transient $m $pn
  grab $m

  set encodings [lsort -dict [encoding names]]

  foreach {wid type title} [list in Input [mc "Input encoding:"] \
      out Output [mc "Output encoding:"]] {
    set w $m.$wid

    labelframe $w -text $title -padx 2m -pady 2m
    listbox $w.list -setgrid yes -width 16 -height 12 \
      -yscroll "$w.scroll set" -exportsel no
    eval "$w.list insert 0 $encodings"
    set index [lsearch -dict -sorted -exact $encodings [get$type\Enc $pn]]
    $w.list selection set $index
    $w.list see $index
    scrollbar $w.scroll -comm "$w.list yview"
    pack $w.scroll -side right -fill y
    pack $w.list -side left
  }

  set dummy [expr {[getInputEnc $pn]==$sysEnc && [getOutputEnc $pn]==$sysEnc}]
  set enableCode "
    if {\${$pn\::dummy}} {
      $m.in.list configure -state disabled
      $m.out.list configure -state disabled
    } else {
      $m.in.list configure -state normal
      $m.out.list configure -state normal
    }
  "
  eval $enableCode
  AmpWidget checkbutton $m.sys -text [
    format [mc "&Use system encoding (%s) for I/O (safest)"] $sysEnc
  ] -var ::tpad::$pn\::dummy -comm [namespace code $enableCode]

  set w $m.but
  frame $w
  AmpWidget button $w.ok -text [mc "&OK"] -width [mcmax &OK &Cancel] -comm \
  [namespace code "
    if {\${$pn\::dummy}} {
      set $pn\::outputEnc \[set $pn\::inputEnc $sysEnc]
    } else {
      set $pn\::inputEnc \[$m.in.list get \[$m.in.list curselection]]
      set $pn\::outputEnc \[$m.out.list get \[$m.out.list curselection]]
    }
    $w.cancel invoke
  "]
  AmpWidget button $w.cancel -text [mc &Cancel] -width [mcmax &OK &Cancel] \
  -comm [namespace code "
    focus $p.text
    after idle {destroy $m}
  "]
  pack $w.ok -padx {0 30} -side left
  pack $w.cancel

  grid $m.in -row 0 -column 0 -padx 10 -pady 10
  grid $m.out -row 0 -column 1 -padx 10 -pady 10
  grid $m.sys -row 1 -column 0 -columnspan 2 -padx 5
  grid $m.but -row 2 -column 0 -columnspan 2 -padx 10 -pady 12

  bind $m <Alt-Key> [list ::tpad::AltKeyInDialog $m %A]
  bind $m <Key-Return> [list ::tk::ButtonInvoke $m.but.ok]
  bind $m <Key-Escape> [list ::tk::ButtonInvoke $m.but.cancel]

  PlaceWindow $pn $m
}

# opens the modal Print setup dialog box
proc ::tpad::printSetup {pn} {
  variable $pn\::p

  set m $p.prn
  toplevel $m
  wm group $m $pn
  wm title $m [mc "Print setup - tpad"]
  wm resizable $m no no
  wm transient $m $pn
  grab $m

  set w $m.what
  frame $w
  bind [AmpWidget label $w.label -text [mc "&Print command"] \
    -width [mcmax "&Print command"] -anchor w] \
    <<AltUnderlined>> [list ::tk::TabToWindow $w.entry]
  entry $w.entry -fg $::tpad::fg -bg $::tpad::bg
  $w.entry delete 0 end
  $w.entry insert 0 $::tpad::lpr
  pack $w.label $w.entry -side left

  set w $m.but
  frame $w
  AmpWidget button $w.ok -text [mc &OK] -width [mcmax &OK &Cancel] -comm \
  [namespace code "
    set lpr \[$m.what.entry get]
    $w.cancel invoke
  "]
  AmpWidget button $w.cancel -text [mc &Cancel] -width [mcmax &OK &Cancel] \
  -comm [namespace code "
    focus $p.text
    after idle {destroy $m}
  "]
  pack $w.ok -padx {0 30} -side left
  pack $w.cancel

  pack $m.what $m.but -padx 10 -pady 12 -anchor w

  bind $m <Alt-Key> [list ::tpad::AltKeyInDialog $m %A]
  bind $m <Key-Return> [list ::tk::ButtonInvoke $m.but.ok]
  bind $m <Key-Escape> [list ::tk::ButtonInvoke $m.but.cancel]

  PlaceWindow $pn $m
  ::tk::TabToWindow $m.what.entry
}

# print the text edited so far
# piping it into lpr(1) or similar
proc ::tpad::print {pn} {
  variable $pn\::p

  set pipeid [open "|$::tpad::lpr" w]
  puts -nonewline $pipeid [$p.text get 1.0 end]
  close $pipeid
}

# quit from from the application
proc ::tpad::quit {pn} {
  variable $pn\::argc
  variable $pn\::next

  if {[OkToContinue $pn]} {
    if {[set files [expr $argc - $next]]>0} {
      switch [tk_messageBox -icon question -title tpad -type yesnocancel \
          -parent $pn -message [format [mc "%d more files to edit.\n\nDo you wish to load the next file in sequence?"] $files]] {
        yes {
	  next $pn
	  return
        }

        no {Quit $pn}

        cancel {return}
      }
    } else {
      Quit $pn
    }
  }
}

# one level undo/redo
proc ::tpad::undoRedo {pn} {
  variable $pn\::p

  if { [catch {$p.text edit undo}] } {
    catch {$p.text edit redo}
  }
  updateStatusBar $pn
}

proc ::tpad::undo {pn} {
  variable $pn\::p

  event generate $p.text <<Undo>>
}

proc ::tpad::redo {pn} {
  variable $pn\::p

  event generate $p.text <<Redo>>
}

proc ::tpad::cut {pn} {
  variable $pn\::p

  event generate $p.text <<Cut>>
  SetMenuOptsState $pn
}

proc ::tpad::copy {pn} {
  variable $pn\::p

  event generate $p.text <<Copy>>
}

proc ::tpad::paste {pn} {
  variable $pn\::p

  # if the text to paste is not empty
  if {![catch {selection get -selection CLIPBOARD}] } {
    # if there is a selection in the text, replace it
    if { ![catch {$p.text index sel.first}] } {
      # turn off automatic insertion of separators
      $p.text configure -autosep no
      $p.text edit separator ;# insert a separator before the replace operation
      $p.text delete sel.first sel.last
    }
    tk_textPaste $p.text
    $p.text edit separator ;# insert a separator after the replace operation
    # turn on automatic insertion of separators
    $p.text configure -autosep yes
    updateStatusBar $pn
    SetMenuOptsState $pn
  }
}

proc ::tpad::tpadDelete {pn} {
  variable $pn\::p

  focus $p.text
  event generate $p.text <Key-Delete>
  updateStatusBar $pn
  SetMenuOptsState $pn
}

# opens the find or find & replace dialog box
# if it isn't already opened
proc ::tpad::findReplace {pn {replace no}} {
  foreach v {p pattern substr} {
    variable $pn\::$v
  }

  set m $p.find
  if { [info commands $m]!={} } {
    wm deiconify $m
    raise $m
  } else {
    toplevel $m
    wm group $m $pn
    wm title $m [expr {$replace ? [mc Replace - tpad"] : [mc "Find - tpad"]}]
    wm resizable $m no no
    wm transient $m $pn

    set w $m.but
    frame $w
    AmpWidget button $w.find -text [mc "Find &Next"] -comm [namespace code "
      set $pn\::pattern \[$m.ww.what.entry get]
      FindPattern $pn
    "]
    AmpWidget button $w.cancel -text [mc &Cancel] -comm [namespace code "
      focus $p.text
      after idle {destroy $m}
    "]
    if {$replace} {
      AmpWidget button $w.replace -text [mc R&eplace] -comm [namespace code "
        set $pn\::pattern \[$m.ww.what.entry get]
        set $pn\::substr \[$m.ww.with.entry get]
        ReplacePattern $pn
      "]
      AmpWidget button $w.all -text [mc "Replace &All"] -comm [namespace code "
        set $pn\::pattern \[$m.ww.what.entry get]
        set $pn\::substr \[$m.ww.with.entry get]
        ReplaceAll $pn
      "]

      pack $w.find $w.replace $w.all $w.cancel -fill x -pady {0 3}
    } else {
      pack $w.find $w.cancel -fill x -pady {0 3}
    }

    frame $m.ww

    set w $m.ww.what
    frame $w
    bind [AmpWidget label $w.label -text [mc &Find:] \
      -width [expr {$replace ? [mcmax &Find: "&Replace with:"] : \
        [mcmax &Find:]}] -anchor w] \
      <<AltUnderlined>> [list ::tk::TabToWindow $w.entry]
    entry $w.entry -width 24 -fg $::tpad::fg -bg $::tpad::bg \
      -validate all -vcmd [namespace code "
        if {{%P}=={{}}} {
          $m.but.find configure -state disabled
          if {$replace} {
            $m.but.replace configure -state disabled
            $m.but.all configure -state disabled
          }
        } else {
          $m.but.find configure -state normal
          if {$replace} {
            $m.but.replace configure -state normal
            $m.but.all configure -state normal
          }
        }
        return yes
      "]
    $w.entry delete 0 end
    $w.entry insert 0 $pattern
    pack $w.label $w.entry -side left

    pack $m.ww.what -pady {0 5}

    if {$replace} {
      set w $m.ww.with
      frame $w
      bind [AmpWidget label $w.label -text [mc "&Replace with:"] -anchor w \
        -width [mcmax &Find: "&Replace with:"]] \
        <<AltUnderlined>> [list ::tk::TabToWindow $w.entry]
      entry $w.entry -width 24 -fg $::tpad::fg -bg $::tpad::bg
      $w.entry delete 0 end
      $w.entry insert 0 $substr
      pack $w.label $w.entry -side left

      pack $m.ww.with
    }

    set w $m.checkbut
    frame $w
    AmpWidget checkbutton $w.match -text [mc "&Match case"] \
      -var ::tpad::$pn\::case -onv yes -offv no
    AmpWidget checkbutton $w.regexp -text [mc "Use rege&xp"] \
      -var ::tpad::$pn\::regexp -onv yes -offv no
    pack $w.match $w.regexp -padx 10 -anchor w

    set w $m.dir
    labelframe $w -text [mc Direction]
    AmpWidget radiobutton $w.up -text [mc &Up] \
      -var ::tpad::$pn\::dir -value up
    AmpWidget radiobutton $w.down -text [mc &Down] \
      -var ::tpad::$pn\::dir -value down
    pack $w.up $w.down -side left

    grid $m.ww -row 0 -column 0 -columnspan 2 -sticky n -padx 5 -pady 10
    grid $m.checkbut -row 1 -column 0 -sticky s -pady 10
    grid $m.dir -row 1 -column 1 -pady 10
    grid $m.but -row 0 -column 2 -rowspan 2 -sticky n -padx 5 -pady 6

    bind $m <Alt-Key> [list ::tpad::AltKeyInDialog $m %A]
    bind $m <Key-Return> [list ::tk::ButtonInvoke $m.but.find]
    bind $m <Key-Escape> [list ::tk::ButtonInvoke $m.but.cancel]

    PlaceWindow $pn $m
  }

  ::tk::TabToWindow $m.ww.what.entry
}

# find the next pattern
# if the pattern is empty (not defined)
# or the Find window is opened, invoke findReplace instead
proc ::tpad::findNext {pn} {
  variable $pn\::p
  variable $pn\::pattern

  if {[info commands $p.find]!={} || $pattern=={}} {
    findReplace $pn
    return
  }
  FindPattern $pn
}

# opens the modal Goto line dialog box
proc ::tpad::goTo {pn} {
  variable $pn\::p

  set m $p.goto
  toplevel $m
  wm group $m $pn
  wm title $m [mc "Goto line - tpad"]
  wm resizable $m no no
  wm transient $m $pn
  grab $m

  set w $m.where
  frame $w
  bind [AmpWidget label $w.label -text [mc "&Line number"] -anchor w \
    -width [mcmax "&Line number"]] \
    <<AltUnderlined>> [list ::tk::TabToWindow $w.spin]
  spinbox $w.spin -width 11 -fg $::tpad::fg -bg $::tpad::bg \
    -from 1 -to [lines $pn] -validate key \
    -vcmd [namespace code {string is integer %P}]
  $w.spin set [lineNo $pn [$p.text index insert]]
  pack $w.label $w.spin -side left

  set w $m.but
  frame $w
  AmpWidget button $w.go -text [mc &Go] -width [mcmax &Go &Cancel] -comm \
  [namespace code "
    if {\[goToLine $pn \[$m.where.spin get]]} {
      $w.cancel invoke
    } else {
      $m.where.spin set \[lineNo $pn \[$p.text index insert]]
      ::tk::TabToWindow $m.where.spin
    }
  "]
  AmpWidget button $w.cancel -text [mc &Cancel] -width [mcmax &Go &Cancel] \
  -comm [namespace code "
    focus $p.text
    after idle {destroy $m}
  "]
  pack $w.go -padx {0 30} -side left
  pack $w.cancel

  pack $m.where $m.but -padx 10 -pady 12 -anchor w

  bind $m <Alt-Key> [list ::tpad::AltKeyInDialog $m %A]
  bind $m <Key-Return> [list ::tk::ButtonInvoke $m.but.go]
  bind $m <Key-Escape> [list ::tk::ButtonInvoke $m.but.cancel]

  PlaceWindow $pn $m
  ::tk::TabToWindow $m.where.spin
}

proc ::tpad::selectAll {pn} {
  variable $pn\::p

  focus $p.text
  event generate $p.text <Control-Key-slash>
}

# insert locale specific date and time
# just before the insertion cursor
proc ::tpad::insertDate {pn} {
  variable $pn\::p

  $p.text insert [$p.text index insert] [clock format [clock seconds] \
    -format $::tpad::timefmt]
  updateStatusBar $pn
}

# insert a file just below the cursor
proc ::tpad::insertFile {pn} {
  variable $pn\::p

  set insFilename [tk_getOpenFile -title [mc "Insert file - tpad"] \
    -filetypes $::tpad::types -parent $pn]

  if {$insFilename!={}} {
    if { [catch {open $insFilename} fileid] } {
      tk_messageBox -title tpad -icon error -message $fileid -parent $pn
    } else {
      fconfigure $fileid -encoding [getInputEnc $pn]

      $p.text insert insert [read $fileid]
      close $fileid
      $p.text see insert

      updateStatusBar $pn
    }
  }
}

proc ::tpad::toggleWordWrap {pn} {
  variable $pn\::p
  variable $pn\::wrap

  $p.text configure -wrap $wrap
  if {$wrap!={none}} {
    grid remove $p.hscroll
  } else {
    grid $p.hscroll
  }
}

proc ::tpad::toggleAsciiTable {pn} {
  variable $pn\::p
  variable $pn\::ascii

  if {$ascii} {
    ascii::display $pn
  } else {
    destroy $p.ascii
  }
}

proc ::tpad::toggleHtmlTidy {pn} {
  variable $pn\::p
  variable $pn\::ht

  if {$::tpad::tidy!={}} {
    if {$ht} {
      ::tpad::tidy::display $pn
    } else {
      destroy $p.tidy
    }
  }
}

proc ::tpad::toggleTextFormatter {pn} {
  variable $pn\::p
  variable $pn\::format

  if {$::tpad::fmtkind!={none}} {
    if {$format} {
      ${::tpad::fmtkind}fmt::display $pn
    } else {
      destroy $p.fmt
    }
  }
}

proc ::tpad::toggleXmlwf {pn} {
  variable $pn\::p
  variable $pn\::wf

  if {$::tpad::xmlwf!={}} {
    if {$wf} {
      ::tpad::xmlwf::display $pn
    } else {
      destroy $p.xmlwf
    }
  }
}

proc ::tpad::toggleStatusBar {pn} {
  variable $pn\::p
  variable $pn\::status

  if {$status} {
    grid $p.statusBar
  } else {
    grid remove $p.statusBar
  }
}

proc ::tpad::tip {pn} {
  variable $pn\::p

  if {$::tpad::fortune!={}} {
    while { [after idle "$p.tip.msg configure -wraplength 6i";
             tk_dialog $p.tip [mc "tpad Tip"] \
             [exec $::tpad::fortune $::tpad::fortunefile] info 1 \
             [mc "Another random tip"] [mc Close]]==0 } {}
  }
}

proc ::tpad::emptyOpenRecent {pn} {
  variable $pn\::p

  if {$::tpad::maxopenrec} {
    set last [expr [$p.menubar.file.recent index last] - 2]
    $p.menubar.file.recent delete 0 [$p.menubar.file.recent index $last]
    $p.menubar.file.recent entryconf end -state disabled
  }
}

proc ::tpad::openRecentLength {pn} {
  variable $pn\::p

  set len [LastOpenRecent $pn]
  if {!$::tpad::tearoff} {
    incr len
  }

  return $len
}

proc ::tpad::help {} {
  eval exec $::tpad::browsercmd $::tpad::helpath/index.html &
}

proc ::tpad::about {pn} {
  tk_messageBox -title [mc "About tpad"] -parent $pn -message [format [mc "%s\n\nCopyright (c) 2004\nAntonio Bonifati\nunder the GNU GPL 2+\n\nant@venus.deis.unical.it"] [whatIs]]
}
## end of public editor menu commands

## init commands
# given pathname (pn), a toplevel window widget created by toplevel(n) or the
# main window (.), builds up an instance of the tpad editor inside it;
# name is the name of the executable (use tpad or tview for the viewer);
# args is a list of options with the same syntax as used on the command line
# (indeed it can be set to $argv or even left empty)
proc ::tpad::tpad {pn name args} {
  namespace eval $pn {}

  variable $pn\::p [expr {$pn=={.} ? {} : $pn}]  ;# pathname prefix
  variable $pn\::argv0 $name
  variable $pn\::argv $args
  variable $pn\::argc [expr {$args=={} ? 0 : [llength $args]}]
  variable $pn\::line {}
  variable $pn\::edit
  variable $pn\::next
  variable $pn\::writable yes

  foreach v {wrap pattern substr dir case regexp status\
             ascii format wf inputEnc outputEnc} {
    variable $pn\::$v [set ::tpad::$v]
  }

  # handle command line parameters (phase 1)
  if {[string match *tview $argv0]} {
    set edit 0
  } else {
    set edit 1
  }

  # two phases Unix-like command line argument parsing

  # phase 1 - parse options
  for {set next 0} {yes} {incr next} {
    set arg [lindex $argv $next]
    set first [string index $arg 0]
    set rest [string range $arg 1 end]

    switch -- $first {
      + {
        set line $rest
      }

      - {
        foreach opt [split $rest {}] {
          switch -- $opt {
            R {
              set edit 0
            }

	    h {
	      cmdHelp
	      exit 1
            }

            default {
              puts [format [mc "%1\$s: illegal option -- %2\$s"] $argv0 $opt]
	      cmdHelp
              exit 1
            }
          }
        }
      }

      default {
        break
      }
    }
  }
}

# start and display a notepad clone application using window pn
# if no additional argument is present, the editor will operate
# on a file (as a standalone and fully independant application)
# and this proc will return nothing, otherwise the additional
# argument must be the name of a variable which will be set upon
# exiting if the user confirmed the editing and leaved unmodified
# otherwise. If setted, the variable will have a copy of the edited
# text. The variable initial value will be used to initialize the
# text to be edited at the beginning. Use trace(n) to execute a
# callback every time the variable $str is modified 
proc ::tpad::display {pn {str {}}} {
  foreach v {argv0 argv p edit line} {
    variable $pn\::$v
  }

  if {$str!={}} {
    variable $pn\::var $str
  } else {
    variable $pn\::var {}
  }

  tk_setPalette $::tpad::palette

  # sets window manager options
  wm title $pn [format $::tpad::titlefmt [file tail $argv0] [mc Untitled]]
  wm iconname $pn [file tail $argv0]
  wm command $pn [concat $argv0 $argv]
  wm group $pn $pn
  # catch the kill of the window manager
  wm protocol $pn WM_DELETE_WINDOW [namespace code "quit $pn"]
  # end of window manager options

  # builds the menus
  set m $p.menubar
  menu $m
  AmpMenu $m add cascade -menu $m.file -lab [mc &File]
  AmpMenu $m add cascade -menu $m.edit -lab [mc &Edit]
  AmpMenu $m add cascade -menu $m.format -lab [mc F&ormat]
  AmpMenu $m add cascade -menu $m.view -lab [mc &View]
  AmpMenu $m add cascade -menu $m.help -lab [mc &Help]

  set w $m.file
  menu $w -tearoff $::tpad::tearoff
  if {$var!={}} {
    AmpMenu $w add comm -lab [mc &Reset] -comm [namespace code "new $pn"]
  } elseif {$edit} {
    AmpMenu $w add comm -lab [mc &New] -accel Ctrl+N \
      -comm [namespace code "new $pn"]
  } else {
    AmpMenu $w add comm -lab [mc &Clear] -accel Ctrl+N \
      -comm [namespace code "new $pn"]
  }
  AmpMenu $w add comm -lab [mc Ne&xt] -comm [namespace code "next $pn"]
  AmpMenu $w add comm -lab [mc &Open...] -accel Ctrl+F12 \
    -comm [namespace code "tpadOpen $pn"]

  if {$::tpad::maxopenrec} {
    AmpMenu $w add cascade -lab [mc "Op&en Recent"] -menu $w.recent
    menu $w.recent -tearoff $::tpad::tearoff
    set entries [LoadOpenRecent $pn]
    $w.recent add separator
    $w.recent add comm -lab [mc "empty this list"] \
      -comm [namespace code "emptyOpenRecent $pn"]
    if {$entries==0} {
      $w.recent entryconf end -state disabled
    }
  }

  if {$edit} {
    if {$var=={}} {  ;# it is safe to remove this test if you want to
      AmpMenu $w add comm -lab [mc &Save] -accel Shift+F12 \
        -comm [namespace code "save $pn"]
    }
    AmpMenu $w add comm -lab [mc "Sa&ve as..."] \
      -comm [namespace code "saveAs $pn"]
  }
  $w add separator
  AmpMenu $w add comm -lab [mc "Pr&int setup..."] \
    -comm [namespace code "printSetup $pn"]
  AmpMenu $w add comm -lab [mc &Print] -accel Ctrl+Shift+F12 \
    -comm [namespace code "print $pn"]
  $w add separator
  AmpMenu $w add comm -lab [mc &Quit] -comm [namespace code "quit $pn"]

  set w $m.edit
  menu $w -tearoff $::tpad::tearoff \
    -postcommand [namespace code "SetStateFromClipboard $pn $w"]
  if {$edit} {
    if {$::tpad::maxundo!=1} {
      AmpMenu $w add comm -lab [mc &Undo] -accel Ctrl+Z \
        -comm [namespace code "undo $pn"]
      AmpMenu $w add comm -lab [mc &Redo] -accel Ctrl+R \
        -comm [namespace code "redo $pn"]
    } else {
      AmpMenu $w add comm -lab [mc &Undo] -accel Ctrl+Z \
        -comm [namespace code "undoRedo $pn"]
    }
    $w add separator
    AmpMenu $w add comm -lab [mc &Cut] -accel Ctrl+X \
      -comm [namespace code "cut $pn"]
  }
  AmpMenu $w add comm -lab [mc C&opy] -accel Ctrl+C \
    -comm [namespace code "copy $pn"]
  if {$edit} {
    AmpMenu $w add comm -lab [mc &Paste] -accel Ctrl+V \
      -comm [namespace code "paste $pn"]
    AmpMenu $w add comm -lab [mc &Delete] -accel Del \
      -comm [namespace code "tpadDelete $pn"]
  }
  $w add separator
  AmpMenu $w add comm -lab [mc &Find...] -accel Ctrl+F \
    -comm [namespace code "findReplace $pn"]
  AmpMenu $w add comm -lab [mc "Find &next"] -accel F3 \
    -comm [namespace code "findNext $pn"]
  if {$edit} {
    AmpMenu $w add comm -lab [mc R&eplace...] -accel Ctrl+H \
      -comm [namespace code "findReplace $pn yes"]
  }
  AmpMenu $w add comm -lab [mc "&Goto line..."] -accel Ctrl+G \
    -comm [namespace code "goTo $pn"]
  $w add separator
  AmpMenu $w add comm -lab [mc "&Select all"] \
    -accel Ctrl+5(kp) -comm [namespace code "selectAll $pn"]
  if {$edit} {
    AmpMenu $w add comm -lab [mc Date/&time] -accel F5 \
      -comm [namespace code "insertDate $pn"]
    AmpMenu $w add comm -lab [mc "&Insert file..."] \
      -comm [namespace code "insertFile $pn"]
  }

  set w $m.format
  menu $w -tearoff $::tpad::tearoff
  AmpMenu $w add checkbutton -var ::tpad::$pn\::wrap \
    -onv word -offv none -lab [mc "&Word wrap"] \
    -comm [namespace code "toggleWordWrap $pn"]
  AmpMenu $w add comm -lab [mc "&I/O Encodings"] \
    -comm [namespace code "ioEncodings $pn"]

  set w $m.view
  menu $w -tearoff $::tpad::tearoff
  AmpMenu $w add checkbutton -var ::tpad::$pn\::status \
    -onv yes -offv no -lab [mc "&Status bar"] \
    -comm [namespace code "toggleStatusBar $pn"]
  $w add separator
  AmpMenu $w add checkbutton -var ::tpad::$pn\::ascii \
    -onv yes -offv no -lab [mc "&Ascii table"] \
    -comm [namespace code "toggleAsciiTable $pn"]
  if {$::tpad::tidy!={}} {
    AmpMenu $w add checkbutton -var ::tpad::$pn\::ht \
      -onv yes -offv no -lab [mc "&HTML Tidy"] \
      -comm [namespace code "toggleHtmlTidy $pn"]
  }
  if {$edit && $::tpad::fmtkind!={none}} {
    AmpMenu $w add checkbutton -var ::tpad::$pn\::format \
      -onv yes -offv no -lab [mc "&Text formatter"] \
      -comm [namespace code "toggleTextFormatter $pn"]
  }
  if {$::tpad::xmlwf!={}} {
    AmpMenu $w add checkbutton -var ::tpad::$pn\::wf \
      -onv yes -offv no -lab [mc "&XML wf checker"] \
      -comm [namespace code "toggleXmlwf $pn"]
  }

  set w $m.help
  menu $w -tearoff $::tpad::tearoff
  AmpMenu $w add comm -lab [mc &Help...] -comm [namespace code help]
  if {$::tpad::fortune!={}} {
    AmpMenu $w add comm -lab [mc &Tips...] -comm [namespace code "tip $pn"]
  }
  AmpMenu $w add comm -lab [mc &About...] -comm [namespace code "about $pn"]

  set w $p.rightmenu
  menu $w -tearoff $::tpad::tearoff \
    -postcommand [namespace code "SetStateFromClipboard $pn $w"]
  if {$edit} {
    if {$::tpad::maxundo!=1} {
      AmpMenu $w add comm -lab [mc &Undo] -comm [namespace code "undo $pn"]
      AmpMenu $w add comm -lab [mc &Redo] -comm [namespace code "redo $pn"]
    } else {
      AmpMenu $w add comm -lab [mc &Undo] \
        -comm [namespace code "undoRedo $pn"]
    }
    $w add separator
    AmpMenu $w add comm -lab [mc &Cut] -comm [namespace code "cut $pn"]
  }
  AmpMenu $w add comm -lab [mc C&opy] -comm [namespace code "copy $pn"]
  if {$edit} {
    AmpMenu $w add comm -lab [mc &Paste] -comm [namespace code "paste $pn"]
    AmpMenu $w add comm -lab [mc &Delete] \
      -comm [namespace code "tpadDelete $pn"]
  }
  $w add separator
  AmpMenu $w add comm -lab [mc "&Select all"] \
    -comm [namespace code "selectAll $pn"]
  # end of menus' definition

  # define the widgets
  set w $p.text
  text $w -font $::tpad::editfont -fg $::tpad::fg -bg $::tpad::bg \
    -wrap none -setgrid yes -undo yes -maxundo $::tpad::maxundo \
    -insertofftime $::tpad::insertofftime -insertontime $::tpad::insertontime \
    -yscroll [namespace code "UpdateOffset $pn"] \
    -xscroll [namespace code "$p.hscroll set"]
  $w tag configure sel -foreground $::tpad::selfg -background $::tpad::selbg
  $w tag configure dupsel -foreground $::tpad::selfg -background $::tpad::selbg
  scrollbar $p.vscroll -comm [namespace code "$w yview"]
  scrollbar $p.hscroll -comm [namespace code "$w xview"] \
    -orient horizontal

  set w $p.statusBar
  frame $w -pady 2
  label $w.filename -anchor w -bd 1 -font $::tpad::statusfont \
    -text [mc Untitled]
  label $w.modified -relief sunken -width [mcmax Modified Readonly] \
    -bd 1 -font $::tpad::statusfont
  label $w.lines -relief sunken -text 1 \
    -width 6 -bd 1 -font $::tpad::statusfont
  label $w.fraction -relief sunken -width 8 -bd 1 -font $::tpad::statusfont
  label $w.coords -relief sunken -text 1,1 \
    -width 8 -bd 1 -font $::tpad::statusfont
  pack $w.filename -side left -exp yes -fill x -padx 2
  pack $w.modified $w.lines $w.fraction $w.coords -side left -padx 2
  # end of widgets' definition

  # install or disable X event handlers
  if {$var=={}} {
    bind $pn <Control-Key-n> [namespace code "new $pn"]
    # this should be preferable than a "bind Text <Control-Key-n> {}"
    bind $p.text <Control-Key-n> [namespace code "new $pn; break"]
  }
  bind $pn <Control-Key-F12> [namespace code "tpadOpen $pn"]
  bind $pn <Control-Shift-Key-F12> [namespace code "print $pn"]
  catch {bind $pn <Control-Shift-Key-XF86_Switch_VT_12> [
    namespace code "print $pn"
  ]}
  catch {bind $p.text <Control-Key-KP_Begin> [namespace code "selectAll $pn"]}
  bind $p.text <Control-Key-Clear> [namespace code "selectAll $pn"]
  bind $p.text <Control-Key-f> [namespace code "findReplace $pn; break"]
  bind $p.text <Control-Key-h> [namespace code "findReplace $pn yes; break"]
  bind $p.text <Control-Key-g> [namespace code "goTo $pn"]
  bind $p.text <Key-F3> [namespace code "findNext $pn"]
  bind $p.text <<Modified>> [namespace code "Modified $pn"]
  bind $p.text <KeyRelease> [namespace code "
    updateStatusBar $pn
    SetMenuOptsState $pn
  "]
  bind $p.text <ButtonRelease> [namespace code "updateStatusBar $pn"]
  bind $p.text <KeyPress> [namespace code "updateStatusBar $pn"]
  bind $p.text <ButtonPress> [namespace code "updateStatusBar $pn"]
  bind $p.text <ButtonPress-3> [namespace code "tk_popup $p.rightmenu %X %Y"]
  bind $p.text <<Selection>> [namespace code "
    DuplicateSelection $pn
    SetMenuOptsState $pn
  "]
  if {$edit} {
    if {$var=={}} {
      bind $pn <Shift-Key-F12> [namespace code "save $pn"]
      catch {bind $pn <Shift-Key-XF86_Switch_VT_12> [namespace code "save $pn"]}
    }
    bind $p.text <Key-F5> [namespace code "insertDate $pn"]
    if {$::tpad::maxundo!=1} {
      event add <<Redo>> <Control-Key-r>
    } else {
      # here it seems better to use break than to destroy the <<Undo>>
      # binding for all Text widgets using "bind Text <<Undo>> {}"
      bind $p.text <<Undo>> [namespace code "undoRedo $pn; break"]
    }
    bind $p.text <<Paste>> [namespace code "paste $pn; break"]
  }
  # end of handlers' definition

  # packing
  $pn config -menu $m
  grid $p.text -row 0 -column 0 -sticky news
  grid $p.vscroll -row 0 -column 1 -sticky ns
  grid $p.hscroll -row 1 -column 0 -sticky ew
  grid $p.statusBar -row 2 -column 0 -columnspan 2 -sticky we
  grid rowconfigure $pn 0 -weight 1
  grid columnconfigure $pn 0 -weight 1
  focus $p.text
  # end of packing

  toggleWordWrap $pn
  toggleTextFormatter $pn
  toggleStatusBar $pn
  toggleXmlwf $pn
  toggleAsciiTable $pn

  # this is phase 2 of command line parameters handling i.e.
  # parse remaining arguments (or multiple file sequential editing)
  if { ![next $pn] } {
    new $pn
  }
  if {$line!={}} {
    goToLine $pn $line
  }

  # make sure to display everything first so the events generated below
  # can't be ignored and the optional modal tip windows
  # always pops up over a complete interface
  update idletasks

  event generate $p.text <<Selection>>
  event generate $p.text <<Modified>>

  if {$::tpad::tip} {tip $pn}

}
## end of init commands

# make sure the msgcat namespace exists
namespace eval ::msgcat {}

namespace eval ::tpad {
  # load language files, stored in msgs subdirectory
  namespace import ::msgcat::mc ::msgcat::mcmax
  ::msgcat::mcload [file join [file dirname [info script]] msgs]

  # load a configuration file (required)
  LoadConf
}
